from typing import Any, List
import numpy as np
import numpy.typing as npt
import pulp


def check_duplicate(alpha_set: npt.NDArray[Any], alpha: npt.NDArray[Any]) -> bool:
    """
    Check whether alpha vector av is already in set a

    :param alpha_set: the set of alpha vectors
    :param alpha: the vector to check
    :return: true or false
    """
    for alpha_i in alpha_set:
        if np.allclose(alpha_i, alpha):
            return True
    return False


def prune_lower_bound(lower_bound: npt.NDArray[Any], S: npt.NDArray[Any]) -> List[Any]:
    """
    Lark's filtering algorithm to prune the lower bound, (Cassandra, Littman, Zhang, 1997)

    :param lower_bound: the current lower bound
    :param S: the set of states
    :return: the pruned lower bound
    """
    # dirty set
    F = set()
    for i in range(len(lower_bound)):
        F.add(tuple(lower_bound[i]))

    # clean set
    Q = []

    for s in S:
        max_alpha_val_s = -np.inf
        max_alpha_vec_s = None
        for alpha_vec in F:
            if alpha_vec[s] > max_alpha_val_s:
                max_alpha_val_s = alpha_vec[s]
                max_alpha_vec_s = alpha_vec
        if max_alpha_vec_s is not None and len(F) > 0:
            # Q.update({max_alpha_vec_s})
            Q.append(np.array(list(max_alpha_vec_s)))
            F.remove(max_alpha_vec_s)
    while F:
        alpha_vec = F.pop()  # just to get a reference
        F.add(alpha_vec)
        x = check_dominance_lp(np.array(alpha_vec), np.array(Q))
        if x is None:
            F.remove(alpha_vec)
        else:
            max_alpha_val = -np.inf
            max_alpha_vec = np.array([])
            for phi in F:
                phi_vec = np.array(list(phi))
                if phi_vec.dot(alpha_vec) > max_alpha_val:
                    max_alpha_val = phi_vec.dot(alpha_vec)
                    max_alpha_vec = phi_vec
            Q.append(max_alpha_vec)
            F.remove(tuple(list(max_alpha_vec)))
    return Q


def check_dominance_lp(alpha_vec: npt.NDArray[Any], Q: npt.NDArray[Any]):
    """
    Uses LP to check whether a given alpha vector is dominated or not (Cassandra, Littman, Zhang, 1997)

    :param alpha_vec: the alpha vector to check
    :param Q: the set of vectors to check dominance against
    :return: None if dominated, otherwise return the vector
    """

    problem = pulp.LpProblem("AlphaDominance", pulp.LpMaximize)

    # --- Decision Variables ----

    # x
    x_vars = []
    for i in range(len(alpha_vec)):
        x_var_i = pulp.LpVariable("x_" + str(i), lowBound=0, upBound=1, cat=pulp.LpContinuous)
        x_vars.append(x_var_i)

    # delta
    delta = pulp.LpVariable("delta", lowBound=None, upBound=None, cat=pulp.LpContinuous)

    # --- Objective Function ----
    problem += delta, "maximize delta"

    # --- The constraints ---

    # x sum to 1
    x_sum = 0
    for i in range(len(x_vars)):
        x_sum += x_vars[i]
    problem += x_sum == 1, "XSumWeights"

    # alpha constraints
    for i, alpha_vec_prime in enumerate(Q):
        x_dot_alpha_sum = 0
        x_dot_alpha_prime_sum = 0
        for j in range(len(alpha_vec)):
            x_dot_alpha_sum += x_vars[j] * alpha_vec[j]
            x_dot_alpha_prime_sum += x_vars[j] * alpha_vec_prime[j]
        problem += x_dot_alpha_sum >= delta + x_dot_alpha_prime_sum, "alpha_constraint _" + str(i)

    # Solve
    problem.solve(pulp.PULP_CBC_CMD(msg=0))

    # Obtain solution
    delta = delta.varValue
    if delta > 0:
        return alpha_vec
    else:
        return None



"""
MIT License

Copyright (c) 2019 CleanRL developers https://github.com/vwxyzjn/cleanrl
"""

from typing import Union, List, Optional, Callable, Tuple, Any
import random
import time
import gymnasium as gym
from gymnasium.wrappers.common import RecordEpisodeStatistics
from gymnasium.spaces.discrete import Discrete
import os
import numpy as np
import torch
import torch.optim as optim
import torch.nn.utils.clip_grad as clip_grad
from torch.distributions.categorical import Categorical
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_common.dao.training.ppo_policy import PPOPolicy
from csle_agents.agents.base.base_agent import BaseAgent
from csle_common.models.ppo_network import PPONetwork
import csle_agents.constants.constants as agents_constants


class PPGCleanAgent(BaseAgent):
    """
    A Phasic Policy Gradient agent using the implementation from CleanRL
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 emulation_env_config: Union[None, EmulationEnvConfig], experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True) -> None:
        """
        Initializes the agent, and sets the hyperparameters as attributes of the class representing the agent.

        :param simulation_env_config: the simulation environment configuration
        :param emulation_env_config: the emulation environment configuration
        :param experiment_config: the experiment configuration
        :param training_job: the training job
        :param save_to_metastore: boolean flag indicating whether the results should be saved to the metastore or not
        """
        super(PPGCleanAgent, self).__init__(simulation_env_config=simulation_env_config,
                                            emulation_env_config=emulation_env_config,
                                            experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.PPG_CLEAN
        self.training_job = training_job
        self.save_to_metastore = save_to_metastore
        config = self.simulation_env_config.simulation_env_input_config
        self.orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)

    def train(self) -> ExperimentExecution:
        """
        Runs the training process

        :return: the results
        """
        pid = os.getpid()

        # Setup experiment metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_TIME_HORIZON)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RANDOM_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNTIME)
        descr = f"Training of policies with Clean-PPG using " \
                f"simulation:{self.simulation_env_config.name}"

        # Setup training job
        if self.training_job is None:
            emulation_name = ""
            if self.emulation_env_config is not None:
                emulation_name = self.emulation_env_config.name
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=emulation_name, simulation_traces=[],
                num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            training_job_id = -1
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
            self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Setup experiment execution
        ts = time.time()
        emulation_name = ""
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        simulation_name = self.simulation_env_config.name
        self.exp_execution = ExperimentExecution(
            result=exp_result, config=self.experiment_config, timestamp=ts,
            emulation_name=emulation_name, simulation_name=simulation_name, descr=descr,
            log_file_path=self.training_job.log_file_path)
        exp_execution_id = -1
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
        self.exp_execution.id = exp_execution_id

        # Training runs, one per seed
        for seed in self.experiment_config.random_seeds:

            # Train
            exp_result, env, model = self.run_ppg(exp_result=exp_result, seed=seed)

            # Save policy
            ts = time.time()
            save_path = f"{self.experiment_config.output_dir}/ppo_policy_seed_{seed}_{ts}.zip"
            model.save(save_path)
            policy = PPOPolicy(
                model=model, simulation_name=self.simulation_env_config.name, save_path=save_path,
                states=self.simulation_env_config.state_space_config.states,
                actions=self.simulation_env_config.joint_action_space_config.action_spaces[
                    self.experiment_config.player_idx].actions, player_type=self.experiment_config.player_type,
                experiment_config=self.experiment_config,
                avg_R=exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN][-1])
            exp_result.policies[seed] = policy

            # Save policy metadata
            if self.save_to_metastore:
                MetastoreFacade.save_ppo_policy(ppo_policy=policy)
                os.chmod(save_path, 0o777)

            # Save trace
            traces = env.get_traces()
            if len(traces) > 0 and self.save_to_metastore:
                MetastoreFacade.save_simulation_trace(traces[-1])
            env.reset_traces()

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                seed_values = []
                for seed_idx in range(len(self.experiment_config.random_seeds)):
                    seed_values.append(value_vectors[seed_idx][i])
                avg_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0])
                std_metrics.append(ExperimentUtil.mean_confidence_interval(
                    data=seed_values,
                    confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1])
            exp_result.avg_metrics[metric] = avg_metrics
            exp_result.std_metrics[metric] = std_metrics
        return self.exp_execution

    def run_ppg(self, exp_result: ExperimentResult, seed: int) -> Tuple[ExperimentResult, BaseEnv, PPONetwork]:
        """
        Runs PPG with a given seed

        :param exp_result: the object to save the experiment results
        :param seed: the random seed
        :return: the updated experiment results, the environment, and the trained model
        """
        Logger.__call__().get_logger().info(f"[CleanPPG] Start training; seed: {seed}")
        clip_coef = self.experiment_config.hparams[agents_constants.PPG_CLEAN.CLIP_COEF].value
        adv_norm_fullbatch = self.experiment_config.hparams[agents_constants.PPG_CLEAN.ADV_NORM_FULLBATCH].value
        clip_vloss = self.experiment_config.hparams[agents_constants.PPG_CLEAN.CLIP_VLOSS].value
        ent_coef = self.experiment_config.hparams[agents_constants.PPG_CLEAN.ENT_COEF].value
        max_grad_norm = self.experiment_config.hparams[agents_constants.PPG_CLEAN.MAX_GRAD_NORM].value
        target_kl = self.experiment_config.hparams[agents_constants.PPG_CLEAN.TARGET_KL].value
        vf_coef = self.experiment_config.hparams[agents_constants.PPG_CLEAN.VF_COEF].value
        learning_rate = self.experiment_config.hparams[agents_constants.PPG_CLEAN.LEARNING_RATE].value
        num_steps = self.experiment_config.hparams[agents_constants.PPG_CLEAN.NUM_STEPS].value
        aux_batch_rollouts = self.experiment_config.hparams[agents_constants.PPG_CLEAN.AUX_BATCH_ROLLOUTS].value
        n_iteration = self.experiment_config.hparams[agents_constants.PPG_CLEAN.N_ITERATION].value
        anneal_lr = self.experiment_config.hparams[agents_constants.PPG_CLEAN.ANNEAL_LR].value
        gamma = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        gae_lambda = self.experiment_config.hparams[agents_constants.PPG_CLEAN.GAE_LAMBDA].value
        e_policy = self.experiment_config.hparams[agents_constants.PPG_CLEAN.E_POLICY].value
        beta_clone = self.experiment_config.hparams[agents_constants.PPG_CLEAN.BETA_CLONE].value
        n_aux_grad_accum = self.experiment_config.hparams[agents_constants.PPG_CLEAN.NUM_AUX_GRAD_ACCUM].value
        num_aux_rollouts = self.experiment_config.hparams[agents_constants.PPG_CLEAN.NUM_AUX_ROLLOUTS].value
        e_auxiliary = self.experiment_config.hparams[agents_constants.PPG_CLEAN.E_AUXILIARY].value
        num_minibatches = self.experiment_config.hparams[agents_constants.PPG_CLEAN.NUM_MINIBATCHES].value
        total_timesteps = self.experiment_config.hparams[agents_constants.PPG_CLEAN.TOTAL_STEPS].value

        envs = gym.vector.SyncVectorEnv([self.make_env() for _ in range(1)])

        # Setup training metrics
        exp_result.all_metrics[seed] = {}
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN] = []
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME] = []
        ExperimentUtil.set_seed(seed)
        cuda = False

        # Create neural network
        device = torch.device(agents_constants.PPO_CLEAN.CUDA if torch.cuda.is_available() and cuda else
                              self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value)
        num_hidden_layers = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS].value
        hidden_layer_dim = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER].value
        input_dim = np.array(envs.single_observation_space.shape).prod()
        env: BaseEnv = self.orig_env
        action_space: Discrete = env.action_space
        action_dim = int(action_space.n)
        agent = PPONetwork(input_dim=input_dim, output_dim_critic=1, output_dim_action=action_dim,
                           num_hidden_layers=num_hidden_layers, hidden_layer_dim=hidden_layer_dim).to(device)
        optimizer = optim.Adam(agent.parameters(), lr=learning_rate, eps=1e-8)

        # seeding
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True

        # Storage setup
        obs = torch.zeros((num_steps, 1) + envs.single_observation_space.shape).to(device) # type: ignore
        actions = torch.zeros((num_steps, 1) + envs.single_action_space.shape).to(device) # type: ignore
        logprobs = torch.zeros((num_steps, 1)).to(device) # type: ignore
        rewards = torch.zeros((num_steps, 1)).to(device) # type: ignore
        horizons = []
        info_returns = []
        dones = torch.zeros((num_steps, 1)).to(device) # type: ignore
        values = torch.zeros((num_steps, 1)).to(device) # type: ignore
        aux_obs = torch.zeros((num_steps, aux_batch_rollouts) + envs.single_observation_space.shape, # type: ignore
                              dtype=torch.uint8)  # type: ignore
        aux_returns = torch.zeros((num_steps, aux_batch_rollouts))

        # Training loop
        global_step = 0
        start_time = time.time()
        next_obs = torch.Tensor(envs.reset()[0]).to(device)
        next_done = torch.zeros(1).to(device)
        batch_size = max(1, int(num_steps))
        num_iterations = max(1, total_timesteps // batch_size)
        minibatch_size = max(2, batch_size // num_minibatches)
        num_phases = max(num_iterations // batch_size, 1)
        aux_batch_rollouts = max(1, n_iteration)

        for phase in range(1, num_phases + 1):

            # POLICY PHASE
            for update in range(1, n_iteration + 1):
                # Annealing the rate if instructed to do so.
                if anneal_lr:
                    frac = 1.0 - (update - 1.0) / num_iterations
                    lrnow = frac * learning_rate
                    optimizer.param_groups[0]["lr"] = lrnow

                for step in range(0, num_steps):
                    global_step += 1 * 1
                    obs[step] = next_obs
                    dones[step] = next_done

                    # action logic
                    with torch.no_grad():
                        action, logprob, _, value = agent.get_action_and_value(next_obs)
                        values[step] = value.flatten()
                    actions[step] = action
                    logprobs[step] = logprob

                    # execute the game and log data.
                    next_obs, reward, done, info, info_d = envs.step(action.cpu().numpy()) # type: ignore
                    if done[0] and "final_info" in info_d:
                        horizons.append(info_d["final_info"][0][agents_constants.ENV_METRICS.TIME_HORIZON])
                        info_returns.append(info_d["final_info"][0][agents_constants.ENV_METRICS.RETURN])
                    rewards[step] = torch.tensor(reward).to(device).view(-1)
                    next_obs, next_done = torch.Tensor(next_obs).to(device), torch.Tensor(done).to(device)

                # bootstrap value if not done
                with torch.no_grad():
                    next_value = agent.get_value(next_obs).reshape(1, -1)
                    advantages = torch.zeros_like(rewards).to(device)
                    lastgaelam = 0
                    for t in reversed(range(num_steps)):
                        if t == num_steps - 1:
                            nextnonterminal = 1.0 - next_done
                            nextvalues = next_value
                        else:
                            nextnonterminal = 1.0 - dones[t + 1]
                            nextvalues = values[t + 1]
                        delta = rewards[t] + gamma * nextvalues * nextnonterminal - values[t]
                        advantages[t] = lastgaelam = delta + gamma * gae_lambda * nextnonterminal * lastgaelam
                    returns = advantages + values

                # flatten the batch
                b_obs = obs.reshape((-1,) + envs.single_observation_space.shape) # type: ignore
                b_logprobs = logprobs.reshape(-1)
                b_actions = actions.reshape((-1,) + envs.single_action_space.shape) # type: ignore
                b_advantages = advantages.reshape(-1)
                b_returns = returns.reshape(-1)
                b_values = values.reshape(-1)

                # PPG code does full batch advantage normalization
                if adv_norm_fullbatch:
                    b_advantages = (b_advantages - b_advantages.mean()) / (b_advantages.std() + 1e-8)

                # Optimizing the policy and value network
                b_inds = np.arange(batch_size)
                clipfracs = []
                for epoch in range(e_policy):
                    np.random.shuffle(b_inds)
                    for start in range(0, batch_size, minibatch_size):
                        end = start + minibatch_size
                        mb_inds = b_inds[start:end]

                        _, newlogprob, entropy, newvalue = agent.get_action_and_value( # type: ignore
                            b_obs[mb_inds], b_actions.long()[mb_inds]) # type: ignore
                        logratio = newlogprob - b_logprobs[mb_inds] # type: ignore
                        ratio = logratio.exp()

                        with torch.no_grad():
                            approx_kl = ((ratio - 1) - logratio).mean()
                            clipfracs += [((ratio - 1.0).abs() > clip_coef).float().mean().item()]

                        mb_advantages = b_advantages[mb_inds] # type: ignore

                        # Policy loss
                        pg_loss1 = -mb_advantages * ratio
                        pg_loss2 = -mb_advantages * torch.clamp(ratio, 1 - clip_coef, 1 + clip_coef)
                        pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                        # Value loss
                        newvalue = newvalue.view(-1)
                        if clip_vloss:
                            v_loss_unclipped = (newvalue - b_returns[mb_inds]) ** 2 # type: ignore
                            v_clipped = b_values[mb_inds] + torch.clamp( # type: ignore
                                newvalue - b_values[mb_inds], -clip_coef, clip_coef) # type: ignore
                            v_loss_clipped = (v_clipped - b_returns[mb_inds]) ** 2 # type: ignore
                            v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped) # type: ignore
                            v_loss = 0.5 * v_loss_max.mean() # type: ignore
                        else:
                            v_loss = 0.5 * ((newvalue - b_returns[mb_inds]) ** 2).mean() # type: ignore

                        entropy_loss = entropy.mean()
                        loss = pg_loss - ent_coef * entropy_loss + v_loss * vf_coef

                        optimizer.zero_grad()
                        loss.backward()
                        clip_grad.clip_grad_norm_(agent.parameters(), max_grad_norm)
                        optimizer.step()

                    if target_kl is not None and approx_kl > target_kl:
                        break

                # PPG Storage - Rollouts are saved without flattening for sampling full rollouts later:
                aux_obs = obs.cpu().clone().to(torch.uint8)
                aux_returns = returns.cpu().clone()

            # AUXILIARY PHASE
            aux_inds = np.arange(aux_batch_rollouts)

            # Build the old policy on the aux buffer before distilling to the network
            aux_pi = torch.zeros((num_steps, aux_batch_rollouts, envs.single_action_space.n)) # type: ignore
            for i, start in enumerate(range(0, aux_batch_rollouts, num_aux_rollouts)):
                end = start + num_aux_rollouts
                m_aux_obs = aux_obs[start:end].to(torch.float32).to(device)
                if len(m_aux_obs) > 0:
                    with torch.no_grad():
                        pi_logits = agent.get_pi(m_aux_obs).logits.cpu().clone()
                    aux_pi[start:end] = pi_logits
                del m_aux_obs

            for auxiliary_update in range(1, e_auxiliary + 1):
                np.random.shuffle(aux_inds)
                for i, start in enumerate(range(0, aux_batch_rollouts, num_aux_rollouts)):
                    end = start + num_aux_rollouts
                    try:
                        m_aux_obs = aux_obs[start:end].to(torch.float32).to(device)
                        if len(m_aux_obs) == 0:
                            continue
                        m_aux_returns = aux_returns[start:end].to(torch.float32).to(device)
                        new_pi, new_values, new_aux_values = agent.get_pi_value_and_aux_value(m_aux_obs)

                        new_values = new_values.view(-1)
                        new_aux_values = new_aux_values.view(-1)
                        old_pi_logits = aux_pi[start:end].to(device)
                        old_pi = Categorical(logits=old_pi_logits)
                        kl_loss = torch.distributions.kl_divergence(old_pi, new_pi).mean()

                        real_value_loss = 0.5 * ((new_values - m_aux_returns) ** 2).mean()
                        aux_value_loss = 0.5 * ((new_aux_values - m_aux_returns) ** 2).mean()
                        joint_loss = aux_value_loss + beta_clone * kl_loss

                        loss = (joint_loss + real_value_loss) / n_aux_grad_accum
                        loss.backward()

                        if (i + 1) % n_aux_grad_accum == 0:
                            clip_grad.clip_grad_norm_(agent.parameters(), max_grad_norm)
                            optimizer.step()
                            optimizer.zero_grad()  # This cannot be outside, else gradients won't accumulate

                    except RuntimeError as e:
                        raise Exception(
                            "if running out of CUDA memory, try a higher --n-aux-grad-accum, which trades more time "
                            "for less gpu memory") from e

                    del m_aux_obs, m_aux_returns

            # Logging
            time_elapsed_minutes = round((time.time() - start_time) / 60, 3)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
            avg_R = round(float(np.mean(info_returns)), 3)
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN].append(round(avg_R, 3))
            avg_T = round(float(np.mean(horizons)), 3)
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON].append(round(avg_T, 3))
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
            running_avg_J = ExperimentUtil.running_average(
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN],
                self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN].append(
                round(running_avg_J, 3))
            running_avg_T = ExperimentUtil.running_average(
                exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON],
                self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_TIME_HORIZON].append(
                round(running_avg_T, 3))
            Logger.__call__().get_logger().info(
                f"[CleanPPG] Iteration: {phase}/{num_phases}, "
                f"avg R: {avg_R}, "
                f"R_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}:"
                f"{running_avg_J}, Avg T:{round(avg_T, 3)}, "
                f"Running_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}_T: "
                f"{round(running_avg_T, 3)}, "
                f"runtime: {time_elapsed_minutes} min")

        envs.close()
        base_env: BaseEnv = envs.envs[0].env.env.env  # type: ignore
        return exp_result, base_env, agent

    def make_env(self) -> Callable[[], RecordEpisodeStatistics[Any, Any]]:
        """
        Helper function for creating the environment to use for training

        :return: a function that creates the environment
        """

        def thunk() -> RecordEpisodeStatistics[Any, Any]:
            """
            Function for creating a new environment

            :return: the created environment
            """
            config = self.simulation_env_config.simulation_env_input_config
            orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)
            env = RecordEpisodeStatistics(orig_env)
            return env

        return thunk

    def hparam_names(self) -> List[str]:
        """
        :return: a list with the hyperparameter names
        """
        return [constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER,
                constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS,
                agents_constants.COMMON.NUM_PARALLEL_ENVS, agents_constants.COMMON.BATCH_SIZE,
                agents_constants.COMMON.EVAL_EVERY, constants.NEURAL_NETWORKS.DEVICE,
                agents_constants.COMMON.SAVE_EVERY, agents_constants.COMMON.NUM_TRAINING_TIMESTEPS,
                agents_constants.PPG_CLEAN.TOTAL_STEPS, agents_constants.PPG_CLEAN.LEARNING_RATE,
                agents_constants.PPG_CLEAN.NUM_STEPS, agents_constants.PPG_CLEAN.ANNEAL_LR,
                agents_constants.PPG_CLEAN.GAMMA, agents_constants.PPG_CLEAN.GAE_LAMBDA,
                agents_constants.PPG_CLEAN.NUM_MINIBATCHES, agents_constants.PPG_CLEAN.ADV_NORM_FULLBATCH,
                agents_constants.PPG_CLEAN.CLIP_COEF, agents_constants.PPG_CLEAN.ENT_COEF,
                agents_constants.PPG_CLEAN.VF_COEF, agents_constants.PPG_CLEAN.MAX_GRAD_NORM,
                agents_constants.PPG_CLEAN.TARGET_KL, agents_constants.PPG_CLEAN.N_ITERATION,
                agents_constants.PPG_CLEAN.E_POLICY, agents_constants.PPG_CLEAN.E_AUXILIARY,
                agents_constants.PPG_CLEAN.BETA_CLONE, agents_constants.PPG_CLEAN.NUM_AUX_ROLLOUTS,
                agents_constants.PPG_CLEAN.NUM_AUX_GRAD_ACCUM, agents_constants.PPG_CLEAN.BATCH_SIZE,
                agents_constants.PPG_CLEAN.MINIBATCH_SIZE, agents_constants.PPG_CLEAN.NUM_ITERATIONS,
                agents_constants.PPG_CLEAN.NUM_PHASES, agents_constants.PPG_CLEAN.AUX_BATCH_ROLLOUTS,
                agents_constants.PPG_CLEAN.V_VALUE]



from typing import List, Optional, Tuple, Any
import math
import time
import os
import numpy as np
import numpy.typing as npt
import gymnasium as gym
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.tabular_policy import TabularPolicy
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_common.util.general_util import GeneralUtil
from csle_agents.agents.base.base_agent import BaseAgent
import csle_agents.constants.constants as agents_constants


class VIAgent(BaseAgent):
    """
    Value Iteration Agent
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True,
                 env: Optional[BaseEnv] = None):
        """
        Initializes the value iteration agent

        :param simulation_env_config: configuration of the simulation environment
        :param experiment_config: the experiment configuration
        :param training_job: an existing training job to use (optional)
        :param save_to_metastore: boolean flag whether to save the execution to the metastore
        :param env: the gym environment for training
        """
        super().__init__(simulation_env_config=simulation_env_config, emulation_env_config=None,
                         experiment_config=experiment_config)
        assert experiment_config.agent_type == AgentType.VALUE_ITERATION
        self.training_job = training_job
        self.save_to_metastore = save_to_metastore
        self.env = env

    def train(self) -> ExperimentExecution:
        """
        Runs the value iteration algorithm to compute V*

        :return: the results
        """
        pid = os.getpid()

        # Initialize metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.VI.DELTA)

        descr = f"Computation of V* with the Value Iteration algorithm using " \
                f"simulation:{self.simulation_env_config.name}"

        for seed in self.experiment_config.random_seeds:
            exp_result.all_metrics[seed] = {}
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []

        if self.env is None:
            self.env = gym.make(self.simulation_env_config.gym_env_name,
                                config=self.simulation_env_config.simulation_env_input_config)

        # Initialize training job
        if self.training_job is None:
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name="", simulation_traces=[],
                num_cached_traces=0,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
                self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Initialize execution result
        ts = time.time()
        emulation_name = None
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        simulation_name = self.simulation_env_config.name
        self.exp_execution = ExperimentExecution(result=exp_result, config=self.experiment_config, timestamp=ts,
                                                 emulation_name=emulation_name, simulation_name=simulation_name,
                                                 descr=descr, log_file_path=self.training_job.log_file_path)
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
            self.exp_execution.id = exp_execution_id

        for seed in self.experiment_config.random_seeds:
            ExperimentUtil.set_seed(seed)
            exp_result = self.value_iteration(exp_result=exp_result, seed=seed)

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                if type(value_vectors[0][0]) is int or type(value_vectors[0][0]) is float \
                        or type(value_vectors[0][0]) is np.int64 or type(value_vectors[0][0]) is np.float64:
                    seed_values = []
                    for seed_idx in range(len(self.experiment_config.random_seeds)):
                        seed_values.append(value_vectors[seed_idx][i])
                    avg = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0]
                    if not math.isnan(avg):
                        avg_metrics.append(avg)
                    ci = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1]
                    if not math.isnan(ci):
                        std_metrics.append(ci)
                    else:
                        std_metrics.append(-1)
                else:
                    avg_metrics.append(-1)
                    std_metrics.append(-1)
                exp_result.avg_metrics[metric] = avg_metrics
                exp_result.std_metrics[metric] = std_metrics

                ts = time.time()
        self.exp_execution.timestamp = ts
        self.exp_execution.result = exp_result
        self.training_job.experiment_result = exp_result
        if self.save_to_metastore:
            MetastoreFacade.update_experiment_execution(experiment_execution=self.exp_execution,
                                                        id=self.exp_execution.id)
            MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
        return self.exp_execution

    def hparam_names(self) -> List[str]:
        """
        :return: a list with the hyperparameter names
        """
        return [agents_constants.COMMON.EVAL_BATCH_SIZE, agents_constants.COMMON.CONFIDENCE_INTERVAL,
                agents_constants.COMMON.RUNNING_AVERAGE, agents_constants.COMMON.GAMMA,
                agents_constants.VI.THETA, agents_constants.VI.TRANSITION_TENSOR,
                agents_constants.VI.REWARD_TENSOR, agents_constants.VI.NUM_STATES, agents_constants.VI.NUM_ACTIONS]

    def value_iteration(self, exp_result: ExperimentResult, seed: int) -> ExperimentResult:
        """
        Runs the value iteration algorithm

        :param exp_result: the experiment result object
        :param seed: the random seed
        :return: the updated experiment result
        """
        theta = self.experiment_config.hparams[agents_constants.VI.THETA].value
        discount_factor = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        num_states = self.experiment_config.hparams[agents_constants.VI.NUM_STATES].value
        num_actions = self.experiment_config.hparams[agents_constants.VI.NUM_ACTIONS].value
        T = self.experiment_config.hparams[agents_constants.VI.TRANSITION_TENSOR].value
        R = self.experiment_config.hparams[agents_constants.VI.REWARD_TENSOR].value
        Logger.__call__().get_logger().info(f"Starting the value iteration algorithm, theta:{theta}, "
                                            f"num_states:{num_states}, discount_factor: {discount_factor}, "
                                            f"num_actions: {num_actions}")
        V, policy, deltas, avg_returns, running_avg_returns = self.vi(
            T=np.array(T), num_states=num_states, num_actions=num_actions,
            R=np.array(R), theta=theta, discount_factor=discount_factor)
        exp_result.all_metrics[seed][agents_constants.VI.DELTA] = deltas
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = avg_returns
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = running_avg_returns
        lookup_table = list(policy)
        for i in range(len(lookup_table)):
            lookup_table[i] = list(lookup_table[i])
        tabular_policy = TabularPolicy(player_type=self.experiment_config.player_type,
                                       actions=self.simulation_env_config.joint_action_space_config.action_spaces[
                                           self.experiment_config.player_idx].actions,
                                       agent_type=self.experiment_config.agent_type, value_function=list(V),
                                       lookup_table=lookup_table, simulation_name=self.simulation_env_config.name,
                                       avg_R=avg_returns[-1])
        exp_result.policies[seed] = tabular_policy
        return exp_result

    def one_step_lookahead(self, state, V, num_actions, num_states, T, discount_factor, R) -> npt.NDArray[Any]:
        """
        Performs a one-step lookahead for value iteration
        :param state: the current state
        :param V: the current value function
        :param num_actions: the number of actions
        :param num_states: the number of states
        :param T: the transition kernel
        :param discount_factor: the discount factor
        :param R: the table with rewards
        :param next_state_lookahead: the next state lookahead table
        :return: an array with lookahead values
        """
        A = np.zeros(num_actions)
        for a in range(num_actions):
            reward = R[a][state]
            for next_state in range(num_states):
                prob = T[a][state][next_state]
                A[a] += prob * (reward + discount_factor * V[next_state])
        return A

    def vi(self, T: npt.NDArray[Any], num_states: int, num_actions: int, R: npt.NDArray[Any],
           theta=0.0001, discount_factor=1.0) \
            -> Tuple[npt.NDArray[Any], npt.NDArray[Any], List[Any], List[Any], List[Any]]:
        """
        An implementation of the Value Iteration algorithm
        :param T: the transition kernel T
        :param num_states: the number of states
        :param num_actions: the number of actions
        :param state_to_id: the state-to-id lookup table
        :param HP: the table with hack probabilities
        :param R: the table with rewards
        :param next_state_lookahead: the next-state-lookahead table
        :param theta: convergence threshold
        :param discount_factor: the discount factor
        :return: (greedy policy, value function, deltas, average_returns)
        """
        deltas = []
        average_returns = []
        running_average_returns = []
        V = np.zeros(num_states)
        iteration = 0
        while True:
            # Stopping condition
            delta = 0
            # Update each state...
            for s in range(num_states):
                # Do a one-step lookahead to find the best action
                A = self.one_step_lookahead(s, V, num_actions, num_states, T, discount_factor, R)
                best_action_value = np.max(A)
                # Calculate delta across all states seen so far
                delta = max(delta, np.abs(best_action_value - V[s]))
                # Update the value function. Ref: Sutton book eq. 4.10.
                V[s] = best_action_value

            deltas.append(delta)

            avg_return = -1.0
            if iteration % self.experiment_config.hparams[agents_constants.COMMON.EVAL_EVERY].value == 0:
                policy = self.create_policy_from_value_function(num_states=num_states, num_actions=num_actions, V=V,
                                                                T=T, discount_factor=discount_factor, R=R)
                avg_return = self.evaluate_policy(policy=policy, eval_batch_size=self.experiment_config.hparams[
                    agents_constants.COMMON.EVAL_BATCH_SIZE].value)
                average_returns.append(avg_return)
                running_avg_J = ExperimentUtil.running_average(
                    average_returns,
                    self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
                running_average_returns.append(running_avg_J)

            if iteration % self.experiment_config.log_every == 0 and iteration > 0:
                Logger.__call__().get_logger().info(f"[VI] i:{iteration}, delta: {delta}, "
                                                    f"theta: {theta}, avg_return: {avg_return}")
            iteration += 1

            # Check if we can stop
            if delta < theta:
                break

        policy = self.create_policy_from_value_function(num_states=num_states, num_actions=num_actions, V=V, T=T,
                                                        discount_factor=discount_factor, R=R)
        return V, policy, deltas, average_returns, running_average_returns

    def evaluate_policy(self, policy: npt.NDArray[Any], eval_batch_size: int) -> float:
        """
        Evalutes a tabular policy

        :param policy: the tabular policy to evaluate
        :param eval_batch_size: the batch size
        :return: None
        """
        if self.env is None:
            raise ValueError("Need to specify an environment to run policy evaluation")
        returns = []
        for i in range(eval_batch_size):
            done = False
            s, _ = self.env.reset()
            R = 0
            while not done:
                if self.simulation_env_config.gym_env_name == "csle-intrusion-response-game-local-pomdp-defender-v1":
                    a = np.random.choice(np.arange(0, len(policy[int(s[0])])), p=policy[int(s[0])])
                    s, r, done, _, info = self.env.step(a)
                    done = True
                else:
                    s, r, done, _, info = self.env.step(policy)
                R += r
            returns.append(R)
        avg_return = np.mean(returns)
        return float(avg_return)

    def create_policy_from_value_function(self, num_states: int, num_actions: int, V: npt.NDArray[Any],
                                          T: npt.NDArray[Any], discount_factor: float, R: npt.NDArray[Any]) \
            -> npt.NDArray[Any]:
        """
        Creates a tabular policy from a value function

        :param num_states: the number of states
        :param num_actions: the number of actions
        :param V: the value function
        :param T: the transition operator
        :param discount_factor: the discount factor
        :param R: the reward function
        :return: the tabular policy
        """
        # Create a deterministic policy using the optimal value function
        policy = np.zeros([num_states, num_actions])
        for s in range(num_states):
            # One step lookahead to find the best action for this state
            A = self.one_step_lookahead(s, V, num_actions, num_states, T, discount_factor, R)
            best_action = np.argmax(A)
            # Always take the best action
            policy[s, best_action] = 1.0
        return policy



from enum import IntEnum


class AcquisitionFunctionType(IntEnum):
    """
    Enum representing different acquisition functions
    """
    EXPECTED_IMPROVEMENT = 0
    CAUSAL_EXPECTED_IMPROVEMENT = 1
    NEGATIVE_LOWER_CONFIDENCE_BOUND = 2
    PROBABILITY_OF_IMPROVEMENT = 3
    PROBABILITY_OF_FEASIBILITY = 4
    ENTROPY_SEARCH = 5
    MAX_VALUE_ENTROPY_SEARCH = 6
    MUMBO = 7



from typing import List, Optional, Tuple, Any
import math
import time
import os
import numpy as np
import numpy.typing as npt
import gymnasium as gym
from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.agent_type import AgentType
from csle_common.util.experiment_util import ExperimentUtil
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.jobs.training_job_config import TrainingJobConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.tabular_policy import TabularPolicy
from csle_common.util.general_util import GeneralUtil
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_agents.agents.base.base_agent import BaseAgent
import csle_agents.constants.constants as agents_constants


class PIAgent(BaseAgent):
    """
    Policy Iteration Agent
    """

    def __init__(self, simulation_env_config: SimulationEnvConfig,
                 experiment_config: ExperimentConfig,
                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True,
                 env: Optional[BaseEnv] = None, create_log_dir: bool = True, env_eval: bool = True,
                 max_eval_length: int = 100, initial_eval_state: int = 0):
        """
        Initializes the value iteration agent

        :param simulation_env_config: configuration of the simulation environment
        :param experiment_config: the experiment configuration
        :param training_job: an existing training job to use (optional)
        :param save_to_metastore: boolean flag whether to save the execution to the metastore
        :param env: the gym environment for training
        :param create_log_dir: Boolean flag whether to create a log directory or not
        :param env_eval: Boolean flag whether to evaluate in a given environment
        :param max_eval_length: maximum evaluation length
        :param initial_eval_state: the initial state for the evaluation
        """
        super().__init__(simulation_env_config=simulation_env_config, emulation_env_config=None,
                         experiment_config=experiment_config, create_log_dir=create_log_dir)
        assert experiment_config.agent_type == AgentType.POLICY_ITERATION
        self.training_job = training_job
        self.save_to_metastore = save_to_metastore
        self.env = env
        self.env_eval = env_eval
        self.max_eval_length = max_eval_length
        self.initial_eval_state = initial_eval_state

    def train(self) -> ExperimentExecution:
        """
        Runs the policy iteration algorithm to compute V*

        :return: the results
        """
        pid = os.getpid()

        # Initialize metrics
        exp_result = ExperimentResult()
        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RETURN)
        exp_result.plot_metrics.append(agents_constants.COMMON.RUNNING_AVERAGE_RETURN)

        descr = f"Computation of V* with the Policy Iteration algorithm using " \
                f"simulation:{self.simulation_env_config.name}"

        for seed in self.experiment_config.random_seeds:
            exp_result.all_metrics[seed] = {}
            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
            exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []

        if self.env is None:
            self.env = gym.make(self.simulation_env_config.gym_env_name,
                                config=self.simulation_env_config.simulation_env_input_config)

        # Initialize training job
        if self.training_job is None:
            self.training_job = TrainingJobConfig(
                simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                progress_percentage=0, pid=pid, experiment_result=exp_result,
                emulation_env_name=None, simulation_traces=[],
                num_cached_traces=0,
                log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                physical_host_ip=GeneralUtil.get_host_ip())
            if self.save_to_metastore:
                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
                self.training_job.id = training_job_id
        else:
            self.training_job.pid = pid
            self.training_job.progress_percentage = 0
            self.training_job.experiment_result = exp_result
            if self.save_to_metastore:
                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)

        # Initialize execution result
        ts = time.time()
        emulation_name = None
        if self.emulation_env_config is not None:
            emulation_name = self.emulation_env_config.name
        simulation_name = self.simulation_env_config.name
        self.exp_execution = ExperimentExecution(result=exp_result, config=self.experiment_config, timestamp=ts,
                                                 emulation_name=emulation_name, simulation_name=simulation_name,
                                                 descr=descr, log_file_path=self.training_job.log_file_path)
        if self.save_to_metastore:
            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
            self.exp_execution.id = exp_execution_id

        for seed in self.experiment_config.random_seeds:
            ExperimentUtil.set_seed(seed)
            exp_result = self.policy_iteration(exp_result=exp_result, seed=seed)

        # Calculate average and std metrics
        exp_result.avg_metrics = {}
        exp_result.std_metrics = {}
        for metric in exp_result.all_metrics[self.experiment_config.random_seeds[0]].keys():
            value_vectors = []
            for seed in self.experiment_config.random_seeds:
                value_vectors.append(exp_result.all_metrics[seed][metric])

            avg_metrics = []
            std_metrics = []
            for i in range(len(value_vectors[0])):
                if type(value_vectors[0][0]) is int or type(value_vectors[0][0]) is float \
                        or type(value_vectors[0][0]) is np.int64 or type(value_vectors[0][0]) is np.float64:
                    seed_values = []
                    for seed_idx in range(len(self.experiment_config.random_seeds)):
                        seed_values.append(value_vectors[seed_idx][i])
                    avg = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[0]
                    if not math.isnan(avg):
                        avg_metrics.append(avg)
                    ci = ExperimentUtil.mean_confidence_interval(
                        data=seed_values,
                        confidence=self.experiment_config.hparams[agents_constants.COMMON.CONFIDENCE_INTERVAL].value)[1]
                    if not math.isnan(ci):
                        std_metrics.append(ci)
                    else:
                        std_metrics.append(-1)
                else:
                    avg_metrics.append(-1)
                    std_metrics.append(-1)
                exp_result.avg_metrics[metric] = avg_metrics
                exp_result.std_metrics[metric] = std_metrics

                ts = time.time()
        self.exp_execution.timestamp = ts
        self.exp_execution.result = exp_result
        self.training_job.experiment_result = exp_result
        if self.save_to_metastore:
            MetastoreFacade.update_experiment_execution(experiment_execution=self.exp_execution,
                                                        id=self.exp_execution.id)
            MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
        return self.exp_execution

    def hparam_names(self) -> List[str]:
        """
        :return: a list with the hyperparameter names
        """
        return [agents_constants.COMMON.EVAL_BATCH_SIZE, agents_constants.COMMON.CONFIDENCE_INTERVAL,
                agents_constants.COMMON.RUNNING_AVERAGE, agents_constants.COMMON.GAMMA,
                agents_constants.VI.THETA, agents_constants.VI.TRANSITION_TENSOR,
                agents_constants.VI.REWARD_TENSOR, agents_constants.VI.NUM_STATES, agents_constants.VI.NUM_ACTIONS]

    def policy_iteration(self, exp_result: ExperimentResult, seed: int) -> ExperimentResult:
        """
        Runs the policy iteration algorithm

        :param exp_result: the experiment result object
        :param seed: the random seed
        :return: the updated experiment result
        """
        N = self.experiment_config.hparams[agents_constants.PI.N].value
        discount_factor = self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value
        num_states = self.experiment_config.hparams[agents_constants.PI.NUM_STATES].value
        num_actions = self.experiment_config.hparams[agents_constants.PI.NUM_ACTIONS].value
        T = self.experiment_config.hparams[agents_constants.PI.TRANSITION_TENSOR].value
        R = self.experiment_config.hparams[agents_constants.PI.REWARD_TENSOR].value
        initial_pi = np.array(self.experiment_config.hparams[agents_constants.PI.INITIAL_POLICY].value)
        Logger.__call__().get_logger().info(f"Starting the policy iteration algorithm, N:{N}, "
                                            f"num_states:{num_states}, discount_factor: {discount_factor}, "
                                            f"num_actions: {num_actions}")
        policy, v, avg_returns, running_avg_returns = self.pi(
            P=np.array(T), num_states=num_states, num_actions=num_actions,
            R=np.array(R), gamma=discount_factor, N=N, policy=initial_pi)
        exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = avg_returns
        exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = running_avg_returns
        tabular_policy = TabularPolicy(player_type=self.experiment_config.player_type,
                                       actions=self.simulation_env_config.joint_action_space_config.action_spaces[
                                           self.experiment_config.player_idx].actions,
                                       agent_type=self.experiment_config.agent_type, value_function=list(v),
                                       lookup_table=list(policy), simulation_name=self.simulation_env_config.name,
                                       avg_R=avg_returns[-1])
        exp_result.policies[seed] = tabular_policy
        return exp_result

    def transition_probability_under_policy(self, P: npt.NDArray[Any], policy: npt.NDArray[Any],
                                            num_states: int) -> npt.NDArray[Any]:
        """
        Utility function for computing the state transition probabilities under the current policy.
        Assumes a deterministic policy (probability 1 of selecting an action in a state)

        Args:
            :P: the state transition probabilities for all actions in the MDP
                (tensor num_actions x num_states x num_states)
            :policy: the policy (matrix num_states x num_actions)
            :num_states: the number of states

        Returns:
               :P_pi: the transition probabilities in the MDP under the given policy
                      (dimensions num_states x num_states)
        """
        P_pi = np.zeros((num_states, num_states))
        for i in range(0, num_states):
            action = np.where(policy[i] == 1)[0][0]
            P_pi[i] = P[action][i]
            assert round(sum(P_pi[i]), 2) == 1  # stochastic
        return P_pi

    def expected_reward_under_policy(self, R: npt.NDArray[Any], policy: npt.NDArray[Any],
                                     num_states: int, num_actions: int) -> npt.NDArray[Any]:
        """
        Utility function for computing the expected immediate reward for each state
        in the MDP given a policy.

        Args:
            :P: the state transition probabilities for all actions in the MDP
                (tensor num_actions x num_states x num_states)
            :policy: the policy (matrix num_states x num_actions)
            :R: the reward function in the MDP (tensor num_actions x num_states x num_states)
            :num_states: the number of states
            :num_actions: the number of actions

        Returns:
               :r: a vector of dimension <num_states> with the expected immediate reward for each state.
        """
        r = np.zeros((num_states))
        for s in range(0, num_states):
            r[s] = sum([policy[s][a] * R[a][s] for a in range(num_actions)])
        return r

    def policy_evaluation(self, P: npt.NDArray[Any], policy: npt.NDArray[Any], R: npt.NDArray[Any], gamma: float,
                          num_states: int, num_actions: int) -> npt.NDArray[Any]:
        """
        Implements the policy evaluation step in the policy iteration dynamic programming algorithm.
        Uses the linear algebra interpretation of policy evaluation, solving it as a linear system.

        Args:
            :P: the state transition probabilities for all actions in the MDP
                (tensor num_actions x num_states x num_states)
            :policy: the policy (matrix num_states x num_actions)
            :gamma: the discount factor
            :num_states: the number of states
            :num_actions: the number of actions

        Returns:
               :v: the state values, a vector of dimension NUM_STATES
        """
        P_pi = self.transition_probability_under_policy(P, policy, num_states=num_states)
        r_pi = self.expected_reward_under_policy(R, policy, num_states=num_states, num_actions=num_actions)
        I = np.identity(num_states)
        v = np.dot(np.linalg.inv(I - (np.dot(gamma, P_pi))), r_pi)
        return np.array(v)

    def policy_improvement(self, P: npt.NDArray[Any], R: npt.NDArray[Any], gamma: float, v: npt.NDArray[Any],
                           pi: npt.NDArray[Any], num_states: int, num_actions: int) -> npt.NDArray[Any]:
        """
        Implements the policy improvement step in the policy iteration dynamic programming algorithm.

        Args:
            :P: the state transition probabilities for all actions in the MDP
                (tensor num_actions x num_states x num_states)
            :R: the reward function in the MDP (tensor num_actions x num_states x num_states)
            :gamma: the discount factor
            :v: the state values (dimension NUM_STATES)
            :pi: the old policy (matrix num_states x num_actions)
            :num_states: the number of states
            :num_actions: the number of actions

        Returns:
               :pi_prime: a new updated policy (dimensions num_states x num_actions)
        """
        pi_prime = np.zeros((num_states, num_actions))
        for s in range(0, num_states):
            action_values = np.zeros(num_actions)
            for a in range(0, num_actions):
                for s_prime in range(0, num_states):
                    action_values[a] += P[a][s][s_prime] * (R[a][s] + gamma * v[s_prime])
            best_action = np.argmax(action_values)
            pi_prime[s][best_action] = 1
        return pi_prime

    def pi(self, P: npt.NDArray[Any], policy: npt.NDArray[Any], N: int, gamma: float, R: npt.NDArray[Any],
           num_states: int, num_actions: int) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], List[float], List[float]]:
        """
        The policy iteration algorithm, interleaves policy evaluation and policy improvement for N iterations.
        Guaranteed to converge to the optimal policy and value function.

        Args:
            :P: the state transition probabilities for all actions in the MDP
            (tensor num_actions x num_states x num_states)
            :policy: the policy (matrix num_states x num_actions)
            :N: the number of iterations (scalar)
            :gamma: the discount factor
            :R: the reward function in the MDP (tensor num_actions x num_states x num_states)

        Returns:
               a tuple of (v, policy) where v is the state values after N iterations and policy is the policy
               after N iterations.
        """
        average_returns = []
        running_average_returns = []
        for i in range(0, N):
            if i % self.experiment_config.hparams[agents_constants.COMMON.EVAL_EVERY].value == 0:
                avg_return = self.evaluate_policy(
                    policy=policy,
                    eval_batch_size=self.experiment_config.hparams[agents_constants.COMMON.EVAL_BATCH_SIZE].value,
                    R=R, P=P, num_states=num_states, num_actions=num_actions
                )
                average_returns.append(avg_return)
                running_avg_J = ExperimentUtil.running_average(
                    average_returns, self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
                running_average_returns.append(running_avg_J)

            if i % self.experiment_config.log_every == 0:
                Logger.__call__().get_logger().info(f"[PI] i:{i}, avg_return: {avg_return}")

            v = self.policy_evaluation(P, policy, R, gamma, num_states=num_states, num_actions=num_actions)
            policy = self.policy_improvement(P, R, gamma, v, policy, num_states=num_states, num_actions=num_actions)

        return policy, v, average_returns, running_average_returns

    def evaluate_policy(self, policy: npt.NDArray[Any], eval_batch_size: int,
                        P: npt.NDArray[Any], R: npt.NDArray[Any], num_states: int, num_actions: int) -> float:
        """
        Evaluates a tabular policy

        :param policy: the tabular policy to evaluate
        :param eval_batch_size: the batch size
        :param P: the transition tensor
        :param R: the reward tensor
        :param num_states: the number of states
        :param num_actions: the number of actions
        :return: the average return of the policy
        """
        returns = []
        for i in range(eval_batch_size):
            done = False
            if self.env is not None and self.env_eval:
                s, _ = self.env.reset()
            else:
                s = self.initial_eval_state
            Ret = 0
            t = 0
            while not done:
                try:
                    a = np.random.choice(np.arange(0, num_actions), p=policy[s])
                except Exception:
                    a = policy
                if self.env is not None and self.env_eval:
                    s, r, done, _, info = self.env.step(a)
                else:
                    r = R[a][s]
                    s = int(np.random.choice(np.arange(0, num_states), p=P[a][s]))
                Ret += r
                t += 1
                if t >= self.max_eval_length:
                    done = True
            returns.append(Ret)
        avg_return = np.mean(returns)
        return float(avg_return)



"""
Constants for csle-agents
"""


class COMMON:
    """
    Common string constants among all agents
    """
    WEIGHTED_INTRUSION_PREDICTION_DISTANCE = "weighted_intrusion_prediction_distance"
    RUNNING_AVERAGE_WEIGHTED_INTRUSION_PREDICTION_DISTANCE = "running_average_weighted_intrusion_prediction_distance"
    START_POINT_CORRECT = "start_point_correct"
    RUNNING_AVERAGE_START_POINT_CORRECT = "running_average_start_point_correct"
    NUM_CACHED_SIMULATION_TRACES = 3
    AVERAGE_RETURN = "average_return"
    RUNNING_AVERAGE_RETURN = "running_average_return"
    AVERAGE_DEFENDER_RETURN = "average_defender_return"
    RUNNING_AVERAGE_DEFENDER_RETURN = "running_average_defender_return"
    AVERAGE_ATTACKER_RETURN = "average_attacker_return"
    RUNNING_AVERAGE_ATTACKER_RETURN = "running_average_attacker_return"
    EVAL_PREFIX = "eval_"
    BASELINE_PREFIX = "baseline_"
    EVAL_EVERY = "eval_every"
    SAVE_EVERY = "save_every"
    EVAL_BATCH_SIZE = "eval_batch_size"
    BATCH_SIZE = "batch_size"
    LEARNING_RATE = "learning_rate"
    GAMMA = "gamma"
    NUM_TRAINING_TIMESTEPS = "num_training_timesteps"
    EXPLOITABILITY = "exploitability"
    POLICY_LOSSES = "policy_losses"
    RUNNING_AVERAGE_EXPLOITABILITY = "running_average_exploitability"
    CONFIDENCE_INTERVAL = "confidence_interval"
    MAX_ENV_STEPS = "max_env_steps"
    RUNNING_AVERAGE = "running_average"
    L = "L"
    NUM_NODES = "num_nodes"
    STOPPING_ENVS = [
        "csle-stopping-game-v1",
        "csle-stopping-game-mdp-attacker-v1",
        "csle-stopping-game-pomdp-defender-v1"
    ]
    RUNNING_AVERAGE_INTRUSION_LENGTH = "running_average_intrusion_length"
    RUNNING_AVERAGE_INTRUSION_START = "running_average_intrusion_start"
    AVERAGE_TIME_HORIZON = "average_time_horizon"
    AVERAGE_UPPER_BOUND_RETURN = "average_upper_bound_return"
    AVERAGE_RANDOM_RETURN = "average_random_return"
    AVERAGE_HEURISTIC_RETURN = "average_heuristic_return"
    RUNNING_AVERAGE_TIME_HORIZON = "running_average_time_horizon"
    NUM_PARALLEL_ENVS = "num_parallel_envs"
    OPTIMIZER = "optimizer"
    ADAM = "Adam"
    SGD = "SGD"
    LEARNING_RATE_EXP_DECAY = "learning_rate_exp_decay"
    LEARNING_RATE_DECAY_RATE = "learning_rate_decay_rate"
    RUNTIME = "runtime"
    EVALUATE_WITH_DISCOUNT = "evaluate_with_discount"
    STATE = "s"
    OBSERVATION = "o"
    REWARD = "r"


class PPO:
    """
    String constants related to PPO
    """
    STEPS_BETWEEN_UPDATES = "steps_between_updates"
    GAE_LAMBDA = "gae_lambda"
    CLIP_RANGE = "clip_range"
    CLIP_RANGE_VF = "clip_range_vf"
    ENT_COEF = "ent_coef"
    VF_COEF = "vf_coef"
    MAX_GRAD_NORM = "max_grad_norm"
    TARGET_KL = "target_kl"
    MLP_POLICY = "MlpPolicy"
    NUM_GRADIENT_STEPS = "num_gradient_steps"


class PPO_CLEAN:
    """
    String constants related to PPO clean
    """
    STEPS_BETWEEN_UPDATES = "steps_between_updates"
    GAE_LAMBDA = "gae_lambda"
    CLIP_RANGE = "clip_range"
    CLIP_RANGE_VF = "clip_range_vf"
    ENT_COEF = "ent_coef"
    VF_COEF = "vf_coef"
    MAX_GRAD_NORM = "max_grad_norm"
    TARGET_KL = "target_kl"
    MLP_POLICY = "MlpPolicy"
    MINIBATCH_SIZE = "minibatch_size"
    NUM_STEPS = "num_steps"
    NUM_MINIBATCHES = "num_minibatches"
    NUM_ENVS = "num_envs"
    ANNEAL_LR = "anneal_lr"
    UPDATE_EPOCHS = "update_epochs"
    CLIP_VLOSS = "clip_vloss"
    REWARD_SCALER = "reward_scaler"
    NORM_ADV = "norm_adv"
    CUDA = "cuda"


class PPG_CLEAN:
    """
    String constants related to PPG clean
    """
    TOTAL_STEPS = "total_steps"
    LEARNING_RATE = "learning_rate"
    NUM_STEPS = "num_steps"
    ANNEAL_LR = "anneal_lr"
    GAMMA = "gamma"
    GAE_LAMBDA = "gae_lambda"
    NUM_MINIBATCHES = "num_minibatches"
    ADV_NORM_FULLBATCH = "adv_norm_fullbatch"
    CLIP_COEF = "clip_coef"
    ENT_COEF = "ent_coef"
    CLIP_VLOSS = "clip_vloss"
    VF_COEF = "vf_coef"
    MAX_GRAD_NORM = "max_grad_norm"
    TARGET_KL = "target_kl"
    N_ITERATION = "n_iteration"
    E_POLICY = "e_policy"
    E_AUXILIARY = "e_auxiliary"
    BETA_CLONE = "beta_clone"
    NUM_AUX_ROLLOUTS = "num_aux_rollouts"
    NUM_AUX_GRAD_ACCUM = "num_aux_grad_accum"
    BATCH_SIZE = "batch_size"
    MINIBATCH_SIZE = "minibatch_size"
    NUM_ITERATIONS = "num_iterations"
    NUM_PHASES = "num_phases"
    AUX_BATCH_ROLLOUTS = "aux_batch_rollouts"
    V_VALUE = "v_value"


class DQN_CLEAN:
    """
    String constants related to the DQN-Clean agent
    """
    STEPS_BETWEEN_UPDATES = "steps_between_updates"
    GAE_LAMBDA = "gae_lambda"
    CLIP_RANGE = "clip_range"
    CLIP_RANGE_VF = "clip_range_vf"
    ENT_COEF = "ent_coef"
    VF_COEF = "vf_coef"
    MAX_GRAD_NORM = "max_grad_norm"
    TARGET_KL = "target_kl"
    MLP_POLICY = "MlpPolicy"
    MINIBATCH_SIZE = "minibatch_size"
    NUM_STEPS = "num_steps"
    NUM_MINIBATCHES = "num_minibatches"
    NUM_ENVS = "num_envs"
    ANNEAL_LR = "anneal_lr"
    UPDATE_EPOCHS = "update_epochs"
    CLIP_VLOSS = "clip_vloss"
    REWARD_SCALER = "reward_scaler"
    NORM_ADV = "norm_adv"
    TAU = "tau"
    EXP_FRAC = "exploration_fraction"
    LEARNING_STARTS = "learning_starts"
    TRAIN_FREQ = "train_frequency"
    T_N_FREQ = "target_network_frequency"
    BUFFER_SIZE = "buffer_size"
    SAVE_MODEL = "save_model"
    CUDA = "cuda"


class C51_CLEAN:
    """
    String constants related to the C51-clean agent
    """
    STEPS_BETWEEN_UPDATES = "steps_between_updates"
    MINIBATCH_SIZE = "minibatch_size"
    NUM_STEPS = "num_steps"
    NUM_ENVS = "num_envs"
    ANNEAL_LR = "anneal_lr"
    CLIP_VLOSS = "clip_vloss"
    REWARD_SCALER = "reward_scaler"
    NORM_ADV = "norm_adv"
    TAU = "tau"
    EXP_FRAC = "exploration_fraction"
    LEARNING_STARTS = "learning_starts"
    TRAIN_FREQ = "train_frequency"
    T_N_FREQ = "target_network_frequency"
    BUFFER_SIZE = "buffer_size"
    START_EXPLORATION_RATE = "start_exploration_rate"
    END_EXPLORATION_RATE = "end_exploration_rate"
    SAVE_MODEL = "save_model"
    CUDA = "cuda"
    N_ATOMS = "n_atoms"
    V_MIN = "v_min"
    V_MAX = "v_max"


class T_FP:
    """
    String constants related to T-FP
    """
    N_2 = "N_2"
    DEFENDER_THRESHOLDS = "defender_thresholds"
    ATTACKER_THRESHOLDS = "attacker_thresholds"
    THETA1_ATTACKER = "theta1_attacker"
    THETA1_DEFENDER = "theta1_defender"
    BEST_RESPONSE_EVALUATION_ITERATIONS = "best_response_evaluation_iterations"
    EQUILIBRIUM_STRATEGIES_EVALUATION_ITERATIONS = "equilibrium_strategies_evaluation_iterations"
    AVERAGE_BEST_RESPONSE_DEFENDER_RETURN = "average_best_response_defender_return"
    RUNNING_AVERAGE_BEST_RESPONSE_DEFENDER_RETURN = "running_average_best_response_defender_return"
    AVERAGE_BEST_RESPONSE_ATTACKER_RETURN = "average_best_response_attacker_return"
    RUNNING_AVERAGE_BEST_RESPONSE_ATTACKER_RETURN = "running_average_best_response_attacker_return"


class VI:
    """
    String constants related to VI
    """
    THETA = "theta"
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    NUM_STATES = "num_states"
    NUM_ACTIONS = "num_actions"
    DELTA = "delta"


class HSVI:
    """
    String constants related to HSVI
    """
    WIDTH = "width"
    WIDTHS = "widths"
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    OBSERVATION_TENSOR = "observation_tensor"
    OBSERVATION_SPACE = "observation_space"
    STATE_SPACE = "state_space"
    ACTION_SPACE = "action_space"
    EPSILON = "epsilon"
    UB_SIZE = "upper_bound_size"
    UB_SIZES = "upper_bound_sizes"
    LB_SIZE = "lower_bound_size"
    LB_SIZES = "lower_bound_sizes"
    INITIAL_BELIEF = "initial_belief"
    USE_LP = "use_lp"
    PRUNE_FREQUENCY = "prune_frequency"
    SIMULATION_FREQUENCY = "simulation_frequency"
    SIMULATE_HORIZON = "SIMULATE_HORIZON"
    NUMBER_OF_SIMULATIONS = "NUMBER_OF_SIMULATIONS"
    INITIAL_BELIEF_VALUES = "initial_belief_values"


class SONDIK_VI:
    """
    String constants related to Sondik's VI
    """
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    OBSERVATION_TENSOR = "observation_tensor"
    OBSERVATION_SPACE = "observation_space"
    STATE_SPACE = "state_space"
    ACTION_SPACE = "action_space"
    INITIAL_BELIEF = "initial_belief"
    USE_PRUNING = "use_pruning"
    PLANNING_HORIZON = "planning_horizon"
    INITIAL_BELIEF_VALUES = "initial_belief_values"
    NUM_ALPHA_VECTORS = "num_alpha_vectors"


class DQN:
    """
    String constants related to DQN
    """
    MLP_POLICY = "MlpPolicy"
    LEARNING_STARTS = "learning_starts"
    EXPLORATION_FRACTION = "exploration_fracion"
    EXPLORATION_INITIAL_EPS = "exploration_initial_eps"
    EXPLORATION_FINAL_EPS = "exploration_final_eps"
    MAX_GRAD_NORM = "max_grad_norm"
    TRAIN_FREQ = "train_freq"
    GRADIENT_STEPS = "gradient_steps"
    N_EPISODES_ROLLOUT = "n_episodes_rollout"
    TARGET_UPDATE_INTERVAL = "target_update_interval"
    BUFFER_SIZE = "buffer_size"
    DQN_BATCH_SIZE = "dqn_batch_size"


class RANDOM_SEARCH:
    """
    String constants related to random search
    """
    L = "L"
    N = "N"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    DELTA = "delta"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"


class SIMULATED_ANNEALING:
    """
    String constants related to simulated annealing
    """
    L = "L"
    N = "N"
    THETA1 = "theta1"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    THETAS = "thetas"
    DELTA = "delta"
    POLICY_TYPE = "policy_type"
    COOLING_FACTOR = "cooling_factor"
    INITIAL_TEMPERATURE = "initial_temperature"
    OBJECTIVE_TYPE = "objective_type"


class NELDER_MEAD:
    """
    String constants related to Nelder-Mead
    """
    L = "L"
    IMPROVE_THRESHOLD = "improve_threshold"
    IMPROVE_BREAK = "improve_break"
    N = "N"
    STEP = "step"
    THETA1 = "theta1"
    UCB = "ucb"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    UTILITY_FUNCTION = "utility_function"
    THETAS = "thetas"
    DELTA = "delta"
    POLICY_TYPE = "policy_type"
    COOLING_FACTOR = "cooling_factor"
    INITIAL_TEMPERATURE = "initial_temperature"
    OBJECTIVE_TYPE = "objective_type"


class BAYESIAN_OPTIMIZATION:
    """
    String constants related to Bayesian Optimization
    """
    L = "L"
    PARAMETER_BOUNDS = "parameter_bounds"
    N = "N"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    UTILITY_FUNCTION = "utility_function"
    UCB_KAPPA = "ucb_kappa"
    UCB_XI = "ucb_xi"
    UCB = "ucb"
    TARGET = "target"
    PARAMS = "params"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"


class CMA_ES_OPTIMIZATION:
    """
    String constants related to CMA-ES Optimization
    """
    L = "L"
    PARAMETER_BOUNDS = "parameter_bounds"
    N = "N"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    UTILITY_FUNCTION = "utility_function"
    UCB_KAPPA = "ucb_kappa"
    UCB_XI = "ucb_xi"
    UCB = "ucb"
    TARGET = "target"
    PARAMS = "params"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"


class PARTICLE_SWARM:
    """
    String constants related to simulated annealing
    """
    S = "S"
    B_LOW = "b_lo"
    B_UP = "b_up"
    INERTIA_WEIGHT = "w"
    COGNITIVE_COEFFICIENT = "Phi_p"
    SOCIAL_COEFFICIENT = "Phi_g"
    L = "L"
    N = "N"
    THETA1 = "theta1"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    THETAS = "thetas"
    DELTA = "delta"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"


class MCS:
    """
    String constants related to Multilevel Coordinate Search
    """
    STEP = "step"
    STEP1 = "step1"
    THETAS = "thetas"
    U = "u"
    V = "v"
    L = "L"
    STOPPING_ACTIONS = "stopping_actions"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    GAMMA = "gamma"
    EPSILON = "epsilon"
    LOCAL = "local"
    IINIT = "iinit"
    M = "m"
    PRT = "prt"
    SMAX = "smax"
    NF = "nf"
    STOP = "stop"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"
    THRESHOLDS = "thresholds"


class BAYESIAN_OPTIMIZATION_EMUKIT:
    """
    String constants related to Bayesian Optimization Emukit
    """
    POLICY_TYPE = "policy_type"
    EVALUATION_BUDGET = "evaluation_budget"
    LENGTHSCALE_RBF_KERNEL = "lengthscale_rbf_kernel"
    VARIANCE_RBF_KERNEL = "variance_rbf_kernel"
    OBS_LIKELIHOOD_VARIANCE = "obs_likelihood_variance"
    BETA = "beta"
    INPUT_SPACE_DIM = "input_space_dim"
    KERNEL_TYPE = "kernel_type"
    ACQUISITION_FUNCTION_TYPE = "acquisition_function_type"
    ACQUISITION_OPTIMIZER_TYPE = "acquisition_optimizer_type"
    X_init = "X_init"
    Y_init = "Y_init"
    PARAMS = "params"
    OBJECTIVE_TYPE = "objective_type"


class DIFFERENTIAL_EVOLUTION:
    """
    String constants related to differential evolution
    """
    L = "L"
    N = "N"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    POPULATION_SIZE = "population_size"
    MUTATE = "mutate"
    RECOMBINATION = "recombination"
    BOUNDS = "bounds"


class CROSS_ENTROPY:
    """
    String constants related to the cross-entropy
    """
    L = "L"
    N = "N"
    K = "K"
    LAMB = "lamb"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"


class KIEFER_WOLFOWITZ:
    """
    String constants related to Kiefer-Wolfowitz
    """
    N = "N"
    L = "L"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    GRADIENT_BATCH_SIZE = "gradient_batch_size"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"
    INITIAL_ALPHA = "initial_alpha"
    DELTA = "delta"
    POLICY_TYPE = "policy_type"


class Q_LEARNING:
    """
    String constants related to Q-learning
    """
    N = "N"
    EPSILON = "epsilon"
    S = "S"
    A = "A"
    INITIAL_STATE_VALUES = "initial_state_values"
    EPSILON_DECAY_RATE = "epsilon_decay_rate"


class SARSA:
    """
    String constants related to Sarsa
    """
    N = "N"
    EPSILON = "epsilon"
    S = "S"
    A = "A"
    INITIAL_STATE_VALUES = "initial_state_values"


class PI:
    """
    String constants related to PI
    """
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    NUM_STATES = "num_states"
    NUM_ACTIONS = "num_actions"
    N = "N"
    INITIAL_POLICY = "initial_policy"


class REINFORCE:
    """
    String constants related to REINFORCE
    """
    N = "N"
    GRADIENT_BATCH_SIZE = "gradient_batch_size"
    CLIP_GRADIENT = "clip_gradient"


class SHAPLEY_ITERATION:
    """
    String constants related to Shapley Iteration
    """
    N = "N"
    DELTA = "delta"
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    STATE_SPACE = "state_space"
    ACTION_SPACE_PLAYER_1 = "action_space_player_1"
    ACTION_SPACE_PLAYER_2 = "action_space_player_2"


class HSVI_OS_POSG:
    """
    String constants related to the HSVI algorithm for OS-POSGs
    """
    N = "N"
    TRANSITION_TENSOR = "transition_tensor"
    REWARD_TENSOR = "reward_tensor"
    STATE_SPACE = "state_space"
    ACTION_SPACE_PLAYER_1 = "action_space_player_1"
    ACTION_SPACE_PLAYER_2 = "action_space_player_2"
    OBSERVATION_SPACE = "observation_space"
    OBSERVATION_FUNCTION = "observation_function"
    INITIAL_BELIEF = "initial_belief"
    PRUNE_FREQUENCY = "prune_frequency"
    EPSILON = "epsilon"
    WIDTHS = "widths"
    EXCESSES = "excesses"


class FICTITIOUS_PLAY:
    """
    String constants related to fictitious play
    """
    PAYOFF_MATRIX = "payoff_matrix"
    N = "N"
    PLAYER_1_PRIOR = "player_1_prior"
    PLAYER_2_PRIOR = "player_2_prior"


class LP_FOR_NF_GAMES:
    """
    String constants related to linear programming for normal-form games
    """
    PAYOFF_MATRIX = "payoff_matrix"
    N = "N"
    ACTION_SPACE_PLAYER_1 = "action_space_player_1"
    ACTION_SPACE_PLAYER_2 = "action_space_player_2"


class LP_FOR_CMDPs:
    """
    String constants related to linear programming for CMDPs
    """
    ACTIONS = "actions"
    STATES = "states"
    COST_TENSOR = "cost_tensor"
    TRANSITION_TENSOR = "transition_tensor"
    CONSTRAINT_COST_TENSORS = "constraint_cost_tensors"
    CONSTRAINT_COST_THRESHOLDS = "constraint_cost_thresholds"


class DYNASEC:
    """
    String constants related to the DynaSec algorithm
    """
    SLEEP_TIME = "sleep_time"
    EMULATION_TRACES_TO_SAVE_W_DATA_COLLECTION_JOB = "emulation_traces_to_save_with_data_collection_job"
    INTRUSION_START_P = "intrusion_start_p"
    TRAINING_EPOCHS = "training_epochs"
    WARMUP_EPISODES = "warmup_episodes"
    NUM_CLIENTS = "num_clients"
    CLIENTS_ARRIVAL_RATE = "clients_arrival_rate"
    EMULATION_MONITOR_SLEEP_TIME = "emulation_monitor_sleep_time"
    STATIC_ATTACKER_TYPE = "static_attacker_type"
    REPLAY_WINDOW_SIZE = "replay_window_size"
    INTRUSION_ALERTS_MEAN = "intrusion_alerts_mean"
    NO_INTRUSION_ALERTS_MEAN = "no_intrusion_alerts_mean"
    INTRUSION_ALERTS_MEAN_BASELINE = "intrusion_alerts_mean_baseline"
    NO_INTRUSION_ALERTS_MEAN_BASELINE = "no_intrusion_alerts_mean_baseline"


class ENV_METRICS:
    """
    String constants representing environment metrics
    """
    RETURN = "R"
    TIME_HORIZON = "T"
    STATE = "s"
    DEFENDER_ACTION = "a1"
    ATTACKER_ACTION = "a2"
    OBSERVATION = "o"
    TIME_STEP = "t"
    AVERAGE_UPPER_BOUND_RETURN = "average_upper_bound_return"
    AVERAGE_RANDOM_RETURN = "average_random_return"
    AVERAGE_HEURISTIC_RETURN = "average_heuristic_return"


class LOCAL_DFSP:
    """
    String constants related to the local DFSP algorithm
    """
    BEST_RESPONSE_EVALUATION_ITERATIONS = "best_response_evaluation_iterations"
    EQUILIBRIUM_STRATEGIES_EVALUATION_ITERATIONS = "equilibrium_strategies_evaluation_iterations"
    AVERAGE_BEST_RESPONSE_DEFENDER_RETURN = "average_best_response_defender_return"
    RUNNING_AVERAGE_BEST_RESPONSE_DEFENDER_RETURN = "running_average_best_response_defender_return"
    AVERAGE_BEST_RESPONSE_ATTACKER_RETURN = "average_best_response_attacker_return"
    RUNNING_AVERAGE_BEST_RESPONSE_ATTACKER_RETURN = "running_average_best_response_attacker_return"
    N_2 = "N_2"


class POMCP:
    """
    String constants related to POMCP
    """
    ROLLOUT_POLICY = "rollout_policy"
    VALUE_FUNCTION = "value_function"
    OBJECTIVE_TYPE = "objective_type"
    S = "S"
    N = "N"
    A = "A"
    O = "O"
    GAMMA = "gamma"
    INITIAL_PARTICLES = "initial_particles"
    REINVIGORATION = "reinvigoration"
    PLANNING_TIME = "planning_time"
    MAX_PARTICLES = "max_particles"
    MAX_NEGATIVE_SAMPLES = "max_negative_samples"
    C = "c"
    C2 = "c2"
    PRIOR_WEIGHT = "prior_weight"
    PRIOR_CONFIDENCE = "prior_confidence"
    MAX_PLANNING_DEPTH = "max_planning_depth"
    USE_ROLLOUT_POLICY = "use_rollout_policy"
    MAX_ROLLOUT_DEPTH = "max_rollout_depth"
    LOG_STEP_FREQUENCY = "log_step_frequency"
    DEFAULT_NODE_VALUE = "default_node_value"
    VERBOSE = "verbose"
    PARALLEL_ROLLOUT = "parallel_rollout"
    NUM_PARALLEL_PROCESSES = "num_parallel_processes"
    NUM_EVALS_PER_PROCESS = "num_evals_per_process"
    ACQUISITION_FUNCTION_TYPE = "acquisition_function_type"
    REINVIGORATED_PARTICLES_RATIO = "reinvigorated_particles_ratio"
    EVAL_ENV_NAME = "eval_env_name"
    EVAL_ENV_CONFIG = "eval_env_config"
    PRUNE_ACTION_SPACE = "prune_action_space"
    PRUNE_SIZE = "prune_size"
